@@ -2,9 +2,6 @@
 Defines L{AccAdapt.Adapter}s for AT-SPI accessibles that manage their 
 descendants.
 
-@todo: PP: add more methods to override DefaultInfo implementations that should
-  return info about items, not the container itself
-
 @author: Peter Parente
 @organization: IBM Corporation
 @copyright: Copyright (c) 2005, 2007 IBM Corporation
@@ -60,6 +57,48 @@ class ContainerAccInfoAdapter(DefaultAccInfoAdapter):
     si = [acc] + [acc.getChildAtIndex(i) for i in xrange(acc.childCount)]
     return si
   
+  @pyLinAcc.errorToLookupError
+  def getAccDefTextAttrs(self):
+    '''
+    Gets all of the default text attributes assigned to the subject if item
+    offset is None, else the default attributes of the item offset child.
+    
+    @return: Name:value pairs of the default text attributes
+    @rtype: dictionary
+    @raise LookupError: When the accessible object is dead
+    @raise IndexError: When the item offset is outside the bounds of the 
+      number of children of the subject accessible
+    @raise NotImplementedError: When this accessible does not support the text
+      interface
+    '''
+    if self.item_offset is None:
+      return super(ContainerAccInfoAdapter, self).getAccDefTextAttrs()
+    c = self.accessible.getChildAtIndex(self.item_offset)
+    if c is None:
+      raise IndexError
+    return IAccessibleInfo(POR(c)).getAccDefTextAttrs()
+    
+  @pyLinAcc.errorToLookupError
+  def getAccAllTextAttrs(self):
+    '''
+    Gets all the text attributes of a given accessible if item offset is None
+    or the attributes of the item offset child if it is not.
+    
+    @return: Name:value pairs of the text attributes at the character offset
+    @rtype: dictionary
+    @raise LookupError: When the accessible object is dead
+    @raise IndexError: When the item offset is outside the bounds of the 
+      number of children of the subject accessible
+    @raise NotImplementedError: When this accessible does not support the text
+      interface
+    '''
+    if self.item_offset is None:
+      return super(ContainerAccInfoAdapter, self).getAccAllTextAttrs()
+    c = self.accessible.getChildAtIndex(self.item_offset)
+    if c is None:
+      raise IndexError
+    return IAccessibleInfo(POR(c)).getAccAllTextAttrs()
+  
   @pyLinAcc.errorToLookupError
   def getAccSelection(self):
     '''  
@@ -94,8 +133,6 @@ class ContainerAccInfoAdapter(DefaultAccInfoAdapter):
       return super(ContainerAccInfoAdapter, self).getAccRoleName()
     try:
       c = self.accessible.getChildAtIndex(self.item_offset)
-      # @todo: PP: just report the parent role for now, otherwise we get 
-      # lots of repeats; revisit
       #l = [i.getLocalizedRoleName() for i in self._getSubItems(c)]
       return unicode(c.getLocalizedRoleName(), 'utf-8')
     except AttributeError:
@@ -335,9 +372,6 @@ class ContainerAccInfoAdapter(DefaultAccInfoAdapter):
     key bindings into tuples since some actions can have multiple ways of 
     activating them dependent on the context.
     
-    @todo: PP: improve so that key bindings that have a ; in them are parsed
-      properly
-    
     @return: List of tuples, each containing one or more string keysym names
       indicating the keys to press in sequence to perform the action
     @rtype: list of tuple of string
@@ -365,8 +399,10 @@ class ContainerAccInfoAdapter(DefaultAccInfoAdapter):
     try:
       c = self.accessible.getChildAtIndex(self.item_offset)
     except AttributeError:
+      # not a valid child
       raise IndexError
-    return IAccessibleInfo(POR(c)).getAccVisualExtents()
+    # keep the character offset into the child
+    return IAccessibleInfo(POR(c,None,self.char_offset)).getAccVisualExtents()
 
   @pyLinAcc.errorToLookupError
   def getAccVisualPoint(self):
@@ -383,8 +419,10 @@ class ContainerAccInfoAdapter(DefaultAccInfoAdapter):
     try:
       c = self.accessible.getChildAtIndex(self.item_offset)
     except AttributeError:
+      # not a valid child
       raise IndexError
-    return IAccessibleInfo(POR(c)).getAccVisualPoint()
+    # keep the character offset into the child
+    return IAccessibleInfo(POR(c,None,self.char_offset)).getAccVisualPoint() 
   
 class ContainerAccActionAdapter(DefaultAccActionAdapter):
   '''
