@@ -43,7 +43,7 @@ def _multiLineCondition(acc):
     # terminals and paragraphs are always multiline
     # note: this does not account for paragraphs that implement hypertext
     return True
-  elif r in (c.ROLE_PAGE_TAB, c.ROLE_LABEL):
+  elif r in (c.ROLE_PAGE_TAB, c.ROLE_LABEL, c.ROLE_EDITBAR):
     # these widgets sometimes show as multiline, but are succinct enough to
     # treat as single line
     return False
@@ -61,7 +61,7 @@ class TextAccInfoAdapter(DefaultAccInfoAdapter):
   as items. Expects the subject to be a L{POR}.
   
   Adapts accessibles that have a role of terminal or have a state of multiline, 
-  single line, or editable and provide the L{Interfaces.IText} interface. 
+  single line, or editable and provide the Text interface. 
   Does not adapt L{POR} accessibles with role of page tab.
   '''
   provides = [IAccessibleInfo]
@@ -162,7 +162,7 @@ class TextNavAdapter(DefaultNavAdapter):
   in the L{IAccessibleNav} interface. 
   
   Adapts accessibles that have a role of terminal or have a state of multiline, 
-  single line, or editable and provide the L{Interfaces.IText} interface. 
+  single line, or editable and provide the Text interface. 
   Does not adapt L{POR} accessibles with role of page tab.
   '''
   provides = [IItemNav, IAccessibleNav]
@@ -377,7 +377,7 @@ class SimpleTextEventHandlerAdapter(DefaultEventHandlerAdapter):
   events for caret and text events. Expects the subject to be a
   L{pyLinAcc.Accessible}.
   
-  Adapts subject accessibles that provide the L{Interfaces.IText} interface.
+  Adapts subject accessibles that provide the Text interface.
   '''
   provides = [IEventHandler]
   
@@ -414,16 +414,17 @@ class SimpleTextEventHandlerAdapter(DefaultEventHandlerAdapter):
     '''
     # get text interface
     text = Interfaces.IText(self.subject)
+    cOff = event.detail1
     # get the text of the caret line, it's starting and ending offsets
     line, sOff, eOff = \
-      text.getTextAtOffset(event.detail1, Constants.TEXT_BOUNDARY_LINE_START)
+      text.getTextAtOffset(cOff, Constants.TEXT_BOUNDARY_LINE_START)
     # create a POR with start offset of the line and relative caret offset
-    por = POR(self.subject, sOff, event.detail1 - sOff)
+    por = POR(self.subject, sOff, cOff - sOff)
 
     # create Caret event with current POR, text added/removed, where it changed,
     # and whether it was inserted.
     return (CaretChange(por, unicode(event.any_data, 'utf-8'), 
-                        event.detail1, (event.type.minor == 'insert'),
+                        cOff, (event.type.minor == 'insert'),
                         **kwargs),)
   
   def _handleCaretEvent(self, event, **kwargs):
@@ -444,7 +445,18 @@ class SimpleTextEventHandlerAdapter(DefaultEventHandlerAdapter):
     text = Interfaces.IText(self.subject)
     # detail1 has new caret position
     caret_offset = event.detail1
-
+    
+    # 2007-05-14
+    # perform a special check account for end of line differences across
+    # implementations; if the offset is greater than the character count
+    # and the last character is not a newline, correct by going back one 
+    # character offset and asking for the line; done to correct problems in
+    # single line entry fields in FF
+    if (caret_offset > 0 and 
+        caret_offset >= text.characterCount and
+        text.getText(caret_offset-1, -1) != '\n'):
+      caret_offset -= 1
+      
     # get the text of the new caret line, it's starting and ending offsets
     line, sOff, eOff = \
       text.getTextAtOffset(caret_offset, Constants.TEXT_BOUNDARY_LINE_START)
@@ -487,6 +499,45 @@ class SimpleTextEventHandlerAdapter(DefaultEventHandlerAdapter):
       sel = text.getText(*text.getSelection(0))
       return (SelectorChange(por, unicode(sel, 'utf-8'), 
                              AEConstants.EVENT_CHANGE_TEXT_SELECT, **kwargs),)
+    
+  def _handleFocusEvent(self, event, **kwargs):
+    '''
+    Creates an L{AEEvent.FocusChange} indicating that the accessible being
+    adapted has gained the focus. Also a L{AEEvent.CaretChange}. This sequence 
+    of L{AEEvent}s will be posted by the caller.
+     
+    @param event: Raw focus change event
+    @type event: L{pyLinAcc.Event.Event}
+    @param kwargs: Parameters to be passed to any created L{AEEvent}
+    @type kwargs: dictionary
+    @return: L{AEEvent.FocusChange} and L{AEEvent.CaretChange}
+    @rtype: tuple of L{AEEvent}
+    '''
+    try:
+      text = Interfaces.IEditableText(self.subject)
+    except NotImplementedError:
+      # only fire caret move when the text is editable, else fire selector
+      return super(SimpleTextEventHandlerAdapter,
+                   self)._handleFocusEvent(event, **kwargs)
+
+    kwargs['focused'] = True
+    # get the text of the caret line, it's starting and ending offsets
+    line = text.getText(0, -1)
+    # create a POR with start offset of the line and relative caret offset
+    por = POR(self.subject, 0, 0)
+    char_offset = text.caretOffset
+    
+    # GTK and Gecko (Firefox) toolkits handle caret events differently.
+    # GTK fails to produce a caret event while Gecko produces one for you.
+    # This code branches on the toolkit type and returns the required events
+    # for each toolkit type.
+    if self.subject.getApplication().toolkitName == 'Gecko':
+      return (FocusChange(por, True, **kwargs),)
+    else:
+      # create a focus event for this POR
+      # and add a caret event to indicate the current position
+      return (FocusChange(por, True, **kwargs),    
+              CaretChange(por, unicode(line, 'utf-8'), char_offset, **kwargs))
 
 class TextEventHandlerAdapter(SimpleTextEventHandlerAdapter):
   '''  
@@ -495,7 +546,7 @@ class TextEventHandlerAdapter(SimpleTextEventHandlerAdapter):
   L{pyLinAcc.Accessible}.
   
   Adapts subject accessibles that have a role of terminal or have a state of 
-  multiline, single line, or editable and provide the L{Interfaces.IText} 
+  multiline, single line, or editable and provide the Text
   interface. Does not adapt accessibles with role of page tab.
   '''
   provides = [IEventHandler]
@@ -538,4 +589,5 @@ class TextEventHandlerAdapter(SimpleTextEventHandlerAdapter):
     # create a focus event for this POR
     # and add a caret event to indicate the current position
     return (FocusChange(por, True, **kwargs),    
-            CaretChange(por, unicode(line, 'utf-8'), sOff, **kwargs))
+            CaretChange(por, unicode(line, 'utf-8'), char_offset-sOff,
+                        **kwargs))
