@@ -5,7 +5,7 @@ response to L{AEEvent}s and L{AEInput.Gesture}s.
 @author: Peter Parente
 @author: Pete Brunet
 @organization: IBM Corporation
-@copyright: Copyright (c) 2005, 2006 IBM Corporation
+@copyright: Copyright (c) 2005, 2007 IBM Corporation
 @license: The BSD License
 
 All rights reserved. This program and the accompanying materials are made 
@@ -13,8 +13,8 @@ available under the terms of the BSD license which accompanies
 this distribution, and is available at
 U{http://www.opensource.org/licenses/bsd-license.php}
 '''
-import logging, weakref, os
-import Task, AEInput, AEEvent, LSRInterfaces, UIElement
+import logging, weakref, os, sys
+import Task, AEInput, AEEvent, LSRInterfaces, UIElement, AEConstants
 from i18n import _
 from AEState import AEState as PerkState
 
@@ -76,8 +76,20 @@ class Perk(Task.Tools.All, UIElement.UIE):
   @type state: L{PerkState}
   @ivar out_caps: Capabilities set by L{setIdealOutput} as ideal for output
     from this L{Perk}. These are used to "late bind" a device for use during
-    execution of this L{Perk} and its L{Task}s.
+    execution of this L{Perk} and its L{Task}s. Defaults to audio.
   @type out_caps: list of string
+  @ivar before_chains: Mapping from target L{Task} name to a list of L{Task}
+    names which should be executed before the target
+  @type before_chains: dictionary of {string : list of string}
+  @ivar after_chains: Mapping from target L{Task} name to a list of L{Task}
+    names which should be executed after the target
+  @type after_chains: dictionary of {string : list of string}
+  @ivar around_chains: Mapping from target L{Task} name to one L{Task} name 
+    which should be executed instead of the target
+  @type around_chains: dictionary of {string : list of string}
+  @ivar registered_modifiers: Mapping from device to a registered modifier for
+  that device
+  @type registered_modifiers: weakref.WeakKeyDictionary
   '''
   STATE = PerkState
   
@@ -95,26 +107,43 @@ class Perk(Task.Tools.All, UIElement.UIE):
     self.keyed_tasks = {}
     self.event_tasks = {}
     self.commands = {}
+    # default to audio output
+    self.out_caps = ['audio']
+    # define empty chains for tasks
+    self.before_chains = {}
+    self.after_chains = {}
+    self.around_chains = {}
+    self.registered_modifiers = weakref.WeakKeyDictionary()
+
     # instantiate an instance of the STATE class; it may be filled with data
     # when preInit is called with state already persisted to disk
     self.state = self.STATE()
-    self.out_caps = []
+    # initialize state settings
+    self.state.init()
     
-  def preInit(self, *args, **kwargs):
-    '''
-    Override pre-initialization to load persisted state from disk.
+  def preInit(self, ae, *args, **kwargs):
+    '''    
+    Overrides pre-initialization to load persisted state from disk. Fetches the
+    managers from the L{AccessEngine} first so that state can be loaded using
+    the L{SettingsManager}. Loads the state, then invokes the super class
+    version of this method to finish initialization.
 
     @note: L{Perk} writers should override L{init} to do initialization, not
-    this method.
+    this method.    
+    @param ae: Reference to AccessEngine for event context
+    @type ae: L{AccessEngine}
     '''
-    # let the tools preInit run first
-    super(Perk, self).preInit(*args, **kwargs)
+    # get managers, this duplicates some work done in the super class, but
+    # it's harmless and only done once per Perk
+    ae.loanManagers(self)
     try:
       # use the settings manager to try to load persisted state for this Perk
       self.state = self.sett_man.loadState(self.getClassName(), self.state)
     except KeyError:
       # do nothing, a blank state object has already been created
       pass
+    # invoke the super class to complete initialization
+    super(Perk, self).preInit(ae, *args, **kwargs)
   
   def postClose(self):
     '''
@@ -132,11 +161,16 @@ class Perk(Task.Tools.All, UIElement.UIE):
     for tasks in (self.named_tasks.values(), self.keyed_tasks.values()):
       map(Task.Task.close, tasks)
       map(Task.Task.postClose, tasks)
+    # unregister modifiers
+    for dev, codes in self.registered_modifiers.iteritems():
+      self.unregisterModifiers(dev, codes)
+      
     # throw away all references
     self.named_tasks = {}
     self.keyed_tasks = {}
     self.event_tasks = {}
     self.commands = {}
+    self.registered_modifiers = weakref.WeakKeyDictionary()
     
   def setIdealOutput(self, *capabilities):
     '''
@@ -173,8 +207,8 @@ class Perk(Task.Tools.All, UIElement.UIE):
 
   def _initTask(self, task):
     '''
-    Initializes a L{Task} by calling L{Task.Task.preInit} and 
-    L{Task.Task.init} if it has not already been initialized.
+    Initializes a L{Task} by calling L{Task.Base.Task.preInit} and 
+    L{Task.Base.Task.init} if it has not already been initialized.
     
     @param task: L{Task} to initialize
     @type task: L{Task}
@@ -196,9 +230,9 @@ class Perk(Task.Tools.All, UIElement.UIE):
     self._initTask(task)
     self.keyed_tasks[tid] = task
     # use the access engine timer to notify our callback
-    self.acc_eng.addTimer(_postTimerEvent, interval*1000, interval, tid, aid,
+    self.acc_eng.addTimer(_postTimerEvent, interval, interval, tid, aid,
                           wtask, weman)
-    self.tier.addPerkRef(tid, self)
+    self.tier.addTaskRef(tid, self)
   
   def registerCommandTask(self, device, codes, name, propogate=False):
     '''
@@ -227,18 +261,23 @@ class Perk(Task.Tools.All, UIElement.UIE):
       # raise a value error if there is already a command registered for this
       # sequence of keys in this Tier
       raise ValueError(
-        'Command with action codes %s on device %s is already registered' % 
+        'command already registered with action codes %s on device %s' % 
         (codes, device.getName()))
     # add key list to our dictionary of commands and store the task to execute
     # as its value
     self.commands[gl] = name
-    self.tier.addPerkRef(gl, self)
+    self.tier.addTaskRef(gl, self)
     if not propogate:
       # try to add a filter to the device, ignore if not supported
       try:
         device.addFilter(gl)
       except NotImplementedError:
         pass
+    # accept a KEY_CMD from the device, ignore if not supported  
+    try:
+      device.addKeyCmd(codes)
+    except NotImplementedError:
+      pass
       
   def registerChooserTask(self, chooser, task):
     '''
@@ -255,7 +294,7 @@ class Perk(Task.Tools.All, UIElement.UIE):
       raise ValueError('Chooser Task already registered')
     self._initTask(task)
     self.keyed_tasks[cid] = task
-    self.tier.addPerkRef(cid, self)
+    self.tier.addTaskRef(cid, self)
     
   def registerNamedTask(self, task, name):
     '''
@@ -277,7 +316,7 @@ class Perk(Task.Tools.All, UIElement.UIE):
       # initialize the Task
       self._initTask(task)
       self.named_tasks[name] = task
-      self.tier.addPerkRef(name, self)
+      self.tier.addTaskRef(name, self)
     else:
       raise ValueError(name)
     
@@ -323,8 +362,9 @@ class Perk(Task.Tools.All, UIElement.UIE):
     '''
     # default to the focus layer if no layer is specified    
     default = not (focus or tier or background)
-    d = {AEEvent.LAYER_FOCUS : focus or default, AEEvent.LAYER_TIER : tier, 
-         AEEvent.LAYER_BACKGROUND : background}
+    d = {AEConstants.LAYER_FOCUS : focus or default, 
+         AEConstants.LAYER_TIER : tier, 
+         AEConstants.LAYER_BACKGROUND : background}
     # get the kind of AEEvent the Task 
     kind = task.getEventType()
     # initialize the Task
@@ -352,9 +392,11 @@ class Perk(Task.Tools.All, UIElement.UIE):
     @raise KeyError: When the L{Task} is not registered
     '''
     tid = id(task)
-    task = self.keyed_tasks[tid]
     del self.keyed_tasks[tid]
-    self.tier.removePerkRef(tid)
+    # let the task finalize itself
+    task.close()
+    task.postClose()
+    self.tier.removeTaskRef(tid)
 
   def unregisterCommandTask(self, device, codes):
     '''
@@ -375,10 +417,16 @@ class Perk(Task.Tools.All, UIElement.UIE):
     # codes are the same
     gl = AEInput.GestureList(device, codes)
     del self.commands[gl]
-    self.tier.removePerkRef(gl)
+    self.tier.removeTaskRef(gl)
     # try to remove a filter from the device, ignore if not supported
     try:
       device.removeFilter(gl)
+    except (NotImplementedError, AttributeError):
+      pass
+    
+    # remove an accepted KEY_CMD from the device, ignore if not supported
+    try:
+      device.removeKeyCmd(codes)
     except NotImplementedError:
       pass
     
@@ -394,7 +442,13 @@ class Perk(Task.Tools.All, UIElement.UIE):
     cid = id(chooser)
     task = self.keyed_tasks[cid]
     del self.keyed_tasks[cid]
-    self.tier.removePerkRef(cid)
+    
+    # We don't want to close the Task because technically it's not getting
+    # unregistered, just decoupled from the chooser it was managing. The Task
+    # must stay registered so that the chooser can be invoked again in the 
+    # future. Do not call Task.close() here!
+    
+    self.tier.removeTaskRef(cid)
 
   def unregisterNamedTask(self, name):
     '''
@@ -410,7 +464,7 @@ class Perk(Task.Tools.All, UIElement.UIE):
     # let the task finalize itself
     task.close()
     task.postClose()
-    self.tier.removePerkRef(name)
+    self.tier.removeTaskRef(name)
     
   def unregisterEventTask(self, task, focus=False, tier=False, 
                           background=False):
@@ -437,8 +491,8 @@ class Perk(Task.Tools.All, UIElement.UIE):
     '''
     # default to the focus layer if no layer is specified    
     default = not (focus or tier or background)
-    d = {AEEvent.LAYER_FOCUS : focus or default, AEEvent.LAYER_TIER : tier, 
-         AEEvent.LAYER_BACKGROUND : background}
+    d = {AEConstants.LAYER_FOCUS : focus or default, AEConstants.LAYER_TIER : tier, 
+         AEConstants.LAYER_BACKGROUND : background}
     # get the task type
     kind = task.getEventType()
     # let the Task finalize itself
@@ -517,11 +571,173 @@ class Perk(Task.Tools.All, UIElement.UIE):
     
     @param event_type: Desired type of L{AEEvent}
     @type event_type: L{AEEvent} class
-    @param task_layer: Layer on which the desired L{Task}s are registered, one 
-      of L{Task.LAYER_FOCUS}, L{Task.LAYER_TIER}, or L{Task.LAYER_BACKGROUND}
+    @param task_layer: Layer on which the desired L{Task}s are registered
     @type task_layer: integer
     @return: List of all L{Task}s that handle the given event type on the given
       layer in this L{Perk}
     @rtype: list of L{Task.Base.Task}
     '''
     return list(self.event_tasks.get((event_type, task_layer), []))
+  
+  def getChainSegment(self, target, link):
+    '''
+    Gets a copy of the segment of the L{Task} chain for target specified by the
+    link type. A copy is returned to avoid iteration errors if the list changes
+    size during execution of the L{Task}s in the chain.
+    
+    @param target: Name of the L{Task} to link to
+    @type target: string
+    @param link: One of the CHAIN constants in L{AEConstants.Tools} 
+    @type link: integer
+    @raise KeyError: When there is no segment for the given target and link
+      type
+    '''
+    if link == AEConstants.CHAIN_BEFORE:
+      return self.before_chains.get(target, [])[:]
+    elif link == AEConstants.CHAIN_AFTER:
+      return self.after_chains.get(target, [])[:]
+    elif link == AEConstants.CHAIN_AROUND:
+      alist = []
+      if self.around_chains.has_key(target):
+        alist.append(self.around_chains[target])
+      return alist
+  
+  def linkTaskToChain(self, name, link, target):
+    '''
+    Links a L{Task} to the one named in target. The L{Task} will be added 
+    either before, after, or around depending on the value of link. Does not
+    allow a L{Task} to link to itself (i.e. name cannot equal target).
+    
+    Invokes L{Tier.Tier.addChainRef} to add a reference to this L{Perk} to the
+    L{Tier} which will later have to traverse the chain across L{Perk}s.
+
+    @param name: Name of the L{Task} to link
+    @type name: string
+    @param link: One of the CHAIN constants in L{AEConstants.Tools} 
+    @type link: integer
+    @param target: Name of the L{Task} to link to
+    @type target: string
+    @raise ValueError: When the link type is unknown
+    @raise AssertionError: When the name equals the target
+    '''
+    assert name != target
+    if link == AEConstants.CHAIN_BEFORE:
+      self.before_chains.setdefault(target, []).append(name)
+    elif link == AEConstants.CHAIN_AFTER:
+      self.after_chains.setdefault(target, []).append(name)
+    elif link == AEConstants.CHAIN_AROUND:
+      self.around_chains[target] = name
+    else:
+      raise ValueError(link)
+    # add a reference to this Perk to the Tier
+    self.tier.addChainRef(target, self)
+      
+  def unlinkTaskFromChain(self, name, link, target):
+    '''
+    Unlinks a L{Task} from the one named in target. The L{Task} will be
+    unlinked either before, after, or around depending on the value of link. 
+    If link is None, the L{Task} will be unlinked from all cases if
+    possible.
+    
+    Invokes L{Tier.Tier.removeChainRef} to delete a reference to this L{Perk}
+    to the L{Tier} which will later have to traverse the chain across L{Perk}s.
+
+    @param name: Name of the L{Task} to link
+    @type name: string
+    @param link: One of the CHAIN constants in L{AEConstants.Tools} 
+    @type link: integer
+    @param target: Name of the L{Task} to unlink from
+    @type target: string
+    @raise ValueError: When the named L{Task} is not unlinked at least once or
+      the link type is unknown
+    @raise KeyError: When the target L{Task} does not have a chain segement in
+      this L{Perk} for the given link type
+    '''
+    if link is not None:
+      if link == AEConstants.CHAIN_BEFORE:
+        self.before_chains[target].remove(name)
+        if len(self.before_chains[target]) == 0:
+          del self.before_chains[target]
+      elif link == AEConstants.CHAIN_AFTER:
+        self.after_chains[target].remove(name)
+        if len(self.after_chains[target]) == 0:
+          del self.after_chains[target]
+      elif link == AEConstants.CHAIN_AROUND:
+        # don't unlink around unless it matches
+        if self.around_chains[target] == name:
+          del self.around_chains[target]
+        else:
+          raise ValueError(name)
+      else:
+        raise ValueError(link)
+    else:
+      # track how many removals we make
+      count = 0
+      # try removing from all chain segments
+      try:
+        self.before_chains[target].remove(name)
+        if len(self.before_chains[target]) == 0:
+          del self.before_chains[target]
+        count += 1
+      except (KeyError, ValueError):
+        pass
+      try:
+        self.after_chains[target].remove(name)
+        if len(self.after_chains[target]) == 0:
+          del self.after_chains[target]
+        count += 1
+      except (KeyError, ValueError):
+        pass
+      try:
+        if self.around_chains[target] == name:
+          # there is only one so get rid of key and value
+          del self.around_chains[target]
+          count += 1
+      except (KeyError, ValueError):
+        pass
+
+      if count == 0:
+        # error if we didn't remove anything
+        raise ValueError
+    # remove reference in tier if they are no longer needed
+    if not self.around_chains.has_key(target) \
+               and not self.before_chains.has_key(target) \
+               and not self.after_chains.has_key(target) :
+      self.tier.removeChainRef(target, self)
+        
+  def getTaskNames(self):
+    '''
+    Gets the names of all L{Task}s registered in this L{Perk}.
+    
+    @return: All names
+    @rtype: list of string
+    '''
+    return self.named_tasks.keys()
+  
+  def unregisterModifiers(self, dev, modifiers):
+    '''
+    Unregisters a list of input modifiers. Ignores the error cases when the
+    device has no registered modifiers or one of the given modifiers is not
+    registered.
+    
+    @param dev: An L{AEInput} device.
+    @type dev: L{AEInput}
+    @param modifiers: A list of key codes used as modifiers.
+    @type modifiers: integer
+    '''
+    for m in modifiers:
+      try:
+        self.registered_modifiers[dev].remove(m)
+      except (KeyError, ValueError):
+        pass
+    
+  def registerModifiers(self, dev, modifiers):
+    '''
+    Registers a list of input modifiers.
+    
+    @param dev: An L{AEInput} device.
+    @type dev: L{AEInput}
+    @param modifiers: A list of key codes used as modifiers.
+    @type modifiers: integer
+    '''
+    self.perk.registered_modifiers.setdefault(dev, []).extend(modifiers)
